import csv
import json
import re
import random
import string
from yaml import Loader, load
from django.conf import settings
from django.core import mail
import os
from django.db import connection
from io import StringIO
from django.core.management import call_command
from django.apps import apps
from types import SimpleNamespace as Namespace
import random
import numpy as np
import requests
import copy
import time
import io




class Constants():
    RESET_SEQ_QUERY = ''
    MODELS = []
    FAIL_LOG = []
    FAILED_TESTS = []
    EXCEPTIONS = []
    PASSED_TESTS = []
    PASSED_LOG = []


class CustomDict(dict):

    def __getattr__(self, key):
        return self[key]

    def __setattr__(self, key, value):
        self[key] = value



def get_random_string(size):
    chars = string.ascii_uppercase + string.digits + string.ascii_lowercase
    return ''.join(random.choice(chars) for _ in range(size))

def add_str_to_each_list_element(str, list_, where='before'):
    if where != 'before':
        new_list = [x + str for x in list_]
    else:
        new_list = [str + x for x in list_]

    return new_list


def get_int_value_if_available(value):
    try:
        match = re.match(r'int\(([0-9]+)\)', str(value))
    except Exception as e:
        return 'lib error: ' + str(e)
    if match is not None:
        return int(match[1])
    else:
        return value

def set_reset_seq_query():
    # global reset_seq_query
    commands = StringIO()

    for app in apps.get_app_configs():
        label = app.label
        call_command('sqlsequencereset', '--no-color', label, stdout=commands)

    Constants.RESET_SEQ_QUERY = commands.getvalue()

def set_all_models():
    # global all_models
    for app in settings.TEST_APPS:
        for model in apps.get_app_config(app).get_models():
            Constants.MODELS.append(model)



def parse_list_string(list_string):
    list_string_ = list_string.strip()
    try:
        float(list_string_)
        return list_string_
    except Exception:
        special_str = '&!$%^*&@~`'
        try:
            list_string_ = list_string_.replace("\\'", special_str)
            list_string_ = list_string_.replace("'", '"')
            list_string_ = list_string_.replace(special_str, "'")
            return json.loads(list_string_)
        except Exception:
            return list_string






def simplify_data(data):
    if data == 'N/A':
        return False

    # if re.match(r'string\((.*)\)', data):
    #     return re.match(r'string\((.*)\)', data)[1]
    if re.match(r'random\((.*, [0-9]+)\)', data):
        matched = re.match(r'random\((.*), ([0-9]+)\)', data)
        random_str = random.choice(matched[1]) * int(matched[2])
        return random_str
    elif re.match(r'integer\(([0-9]+)\)', data):
        return int(re.match(r'integer\(([0-9]+)\)', data)[1])
    elif re.match(r'bool\((.*)\)', data):
        return BOOL.get(re.match(r'bool\((.*)\)', data)[1])
    elif re.match(r'attachment\((.*)\)', data):
        attachment_file = re.match('attachment\((.*)\)', data)[1]
        return open(settings.TEST_DATA_PATH + '/upload/' + attachment_file, 'rb')
    # elif re.match(r'Context\.(.*)', data):
    #     context_variable = re.match(r'Context\.(.*)', data)[1]
    #     return Context.get(context_variable)
    else:
        return parse_list_string(data)



def request_response_formatter(file):
    with open(settings.TEST_DATA_PATH + '/' + file, 'r') as test_file:
        test_data = csv.reader(test_file, delimiter=',', quotechar='"')
        header = next(test_data)
        all_req_resp = []

        for row in test_data:
            req = {'files': {}}
            resp = {}
            headers = {}

            for index, each_header in enumerate(header):
                simplified_data = simplify_data(row[index])
                parsed_obj = parse_snapshot(simplified_data)
                if simplified_data:
                    if each_header[:5] == 'resp_':
                        resp[each_header[5:]] = simplified_data
                    elif each_header[:7] == 'header_':
                        if each_header[7:] == 'headers' and type(parsed_obj) is dict:
                            headers.update(parsed_obj)
                        else:
                            headers[each_header[7:]] = parsed_obj
                    elif isinstance(parsed_obj, io.IOBase):
                        files = req.get('files')
                        files.update({
                            each_header: parsed_obj
                        })
                        req['files'] = files
                    else:
                        if type(parsed_obj) is dict and each_header == 'request_body':
                            req.update(parsed_obj)
                        else:
                            req[each_header] = parsed_obj
            

            result = {'req': req, 'resp': resp, 'headers': headers}
            all_req_resp.append(result)

    return all_req_resp

def get_all_locales():
    all_locales = []
    locales_dir = settings.LOCALE_DIR
    if os.path.isdir(locales_dir):
        all_locales += os.listdir(locales_dir)
    return all_locales

def reset_db(*arg):
    for model in Constants.MODELS:
        model.objects.all().delete()
    cursor = connection.cursor()
    cursor.execute(Constants.RESET_SEQ_QUERY)

def get_translations(file):
    with open(
            settings.TEST_PAYLOAD_PATH + '/translation/' + file + '.yml',
            'r',
            encoding='utf8') as fp:
        translations = load(fp, Loader)
    return translations

BOOL = {
    'true': True,
    'True': True,
    'false': False,
    'False': False,
    'None': False,
    'none': False
}

def get_data_from_yml(file):
    with open(settings.TEST_DATA_PATH + '/yml/' + file, encoding='utf8') as fp:
        data = load(fp,Loader)
    return data


def set_auth_header(client, auth_token):
    existing_headers = client.headers
    existing_headers['Authorization'] = auth_token
    client.headers.update(existing_headers)

def reset_auth_header(client):
    existing_headers = client.headers
    if existing_headers.get('Authorization'):
        existing_headers.pop('Authorization')
    client.headers.update(existing_headers)

def set_lang_header(client, lang):
    existing_headers = client.headers
    existing_headers['Accept-Language'] = lang
    client.headers.update(existing_headers)

def set_custom_headers(client, headers):
    existing_headers = client.headers
    existing_headers = {**existing_headers, **headers}
    client.headers.update(existing_headers)


def format_response(response_obj):
    try:
        response = json.loads(response_obj.content)
    except Exception:
        response = {}
    if type(response) is list:
        response = {'content': response}

    response['status_code'] = str(response_obj.status_code)
    return response

def process_request_response(**kwargs):
    files = None
    client = kwargs.get('client')
    
    if kwargs.get('data'):
        files = kwargs.get('data').pop('files')

    data = kwargs.get('data', {})

    processor = getattr(client, kwargs['method'])
    if files:
        response = processor(
            kwargs['url'],
            data=data,
            files=files
            )
    else:
        response = processor(
            kwargs['url'],
            json=data
            )
    
    for k,v in files.items():
        data[k] = str(v)

    response_time = response.elapsed.total_seconds() * 1000
    response = format_response(response)
    response['response_time'] = float("%0.2f"%(response_time))
    return response

def unit_test_formatter(file_name):
    with open(settings.TEST_DATA_PATH + '/unittest/' + file_name, 'r')\
            as test_file:
        test_data = csv.reader(test_file, delimiter=',', quotechar='"')
        header = next(test_data)

        all_test_data = []
        args_idx = []
        kwargs_idx = []
        inits_idx = []
        returns_idx = None

        for index, each_header in enumerate(header):
            if each_header[:7] == 'kwargs_':
                kwargs_idx.append(index)
            elif each_header[:5] == 'init_':
                inits_idx.append(index)
            elif each_header == 'returns':
                returns_idx = index
            else:
                args_idx.append(index)

        for each_row in test_data:
            inits = []
            args = []
            kwargs = {}
            returns = ''

            for init_idx in inits_idx:
                inits.append(get_int_value_if_available(each_row[init_idx]))

            for arg_idx in args_idx:
                args.append(get_int_value_if_available(each_row[arg_idx]))

            for kwarg_idx in kwargs_idx:
                kwargs[header[kwarg_idx][7:]] = get_int_value_if_available(
                    each_row[kwarg_idx])

            returns = get_int_value_if_available(each_row[returns_idx])

            all_test_data.append({
                'inits': inits,
                'args': args,
                'kwargs': kwargs,
                'returns': returns
            })

        return all_test_data

def dict_to_obj(resp):
    try:
        if type(resp) is dict:
            resp = json.dumps(resp)
        else:
            resp = str(resp, 'UTF-8')
        return json.loads(resp, object_hook=lambda d: CustomDict(d))
    except Exception as e:
        return json.loads(
            json.dumps({
                'error': resp
            }), object_hook=lambda d: CustomDict(d))

def get_test_endpoints(file):
    with open(settings.TEST_PAYLOAD_PATH + '/' + file, 'r') as fp:
        endpoints = load(fp, Loader)

    for k, v in endpoints.items():
        endpoints[k] = settings.TEST_SERVER + v
    endpoints = dict_to_obj(endpoints)
    return endpoints




def generate_test_report(**kwargs):
    # kwargs = dict_to_obj(kwargs)
    report_ = [
                kwargs.get('test_name'), kwargs.get('test_id'), 
                kwargs.get('priority'), kwargs.get('purpose'), 
                kwargs.get('reproduce_steps')
            ]
    report = {
        'test_id': kwargs.get('test_id'),
        'test': kwargs.get('purpose'),
        'reproduce_steps': kwargs.get('reproduce_steps'),
        'request_body': json.dumps(kwargs.get('request_body'), indent=4, sort_keys=False, ensure_ascii=False),
        'request_header': json.dumps(kwargs.get('request_header'), indent=4, sort_keys=False, ensure_ascii=False),
        'response': json.dumps(kwargs.get('response'), indent=4, sort_keys=False, ensure_ascii=False),
        'expected_response': json.dumps(kwargs.get('expected_response'), indent=4, sort_keys=False, ensure_ascii=False),
        'error_info': kwargs.get('error_info'),
        'response_time': kwargs.get('response_time'),
    }
    Constants.PASSED_LOG.append(report_)
    Constants.PASSED_TESTS.append(report)

def mark_test_as_failed():
    Constants.FAIL_LOG.append(Constants.PASSED_LOG[-1])
    Constants.FAILED_TESTS.append(Constants.PASSED_TESTS[-1])
    del Constants.PASSED_TESTS[-1]
    del Constants.PASSED_LOG[-1]
    # Constants.FAILED_TESTS[-1]['failed'] = True


def generate_analytics(fail_log):
    if not fail_log:
        return []
    Constants.FAIL_LOG = np.array(fail_log)
    priorities, counts = np.unique(Constants.FAIL_LOG[:, 2], return_counts=True)
    # priorities = [str(x) for x in priorities]
    prior_count = list(zip(counts, priorities))
    # return [' '.join(x) for x in prior_count]
    return prior_count


def replace_attribute_value(obj, attributes):
    key_val = attributes.split('=')
    keys = key_val[0].split('.')
    value = simplify_data(key_val[1])
    temp = [obj]
    for index, key in enumerate(keys):
        temp.append(temp[index].get(key))
    temp[-1] = value
    count = len(temp) - 1
    while(count):
        temp[count-1][keys[count-1]] = temp[count]
        count-=1
    return temp[0]




def parse_snapshot(snapshot, actual=None):
    matched = re.match(r'snapshot\((.*\.json)\)((\.)\((.*)\))?', str(snapshot))
    if matched:
        snapshot_file = matched[1]
        
        if os.path.isfile(
            settings.TEST_DATA_PATH + '/snapshots/' + snapshot_file):

            f = open(settings.TEST_DATA_PATH + '/snapshots/' +
                                snapshot_file, 'r')

            try:
                obj = json.load(f)
            except Exception:
                return f.read()

            if matched[4]:              
                x = matched[4].split(',')
                for _ in x:
                    obj = replace_attribute_value(obj, _)
                return obj
            else:
                return obj
        else:
            if actual:
                f = open(settings.TEST_DATA_PATH + '/snapshots/' +
                                snapshot_file, 'w', encoding='utf-8')
            
                json.dump(actual, f, ensure_ascii=False, indent=4)
                return actual     

    return snapshot


def get_attribute_value_from_json(json_obj, attribute_path):
    if attribute_path and type(json_obj) is dict:
        attribute_value = copy.deepcopy(json_obj)
        attribute_path = attribute_path.split('.')
        for attribute in attribute_path:
            list_available = re.match(r'^(.*)\[([0-9]+)\]$', attribute)
            index = None
            if list_available:
                attribute = list_available[1]
                index = int(list_available[2])

            if type(attribute_value) is dict:
                attribute_value = attribute_value.get(attribute)
                if index is not None:
                    attribute_value = attribute_value[index]
            else:
                return attribute_path
                # raise Exception('attribute not found in dictionary object')
        
        return attribute_value

    raise Exception('get_attribute_value_from_json: argument error')



# expects csv file with base_url alias, endpoint, headers & req body
def create_default_data_by_api(file):
    data = request_response_formatter('default/' + file)
    
    for _ in data:
        url = getattr(settings, _.get('req').pop('base_url').upper() + '_BASE_URL') + \
            _.get('req').pop('endpoint')
        request_method = _.get('req').get('method', 'post')
        request_body = _.get('req')
        client = requests.Session()
        headers = _.get('headers')
        
        
        request_body = replace_context_var(request_body)
        client.headers.update(replace_context_var(headers))

        response = process_request_response(
            client=client,
            url=form_endpoint(url, request_body),
            data=request_body,
            method=request_method
        )

        if int(response.get('status_code')) in range(200, 205):
            print('Default Data Created with ' + url)
            if _.get('req').get('set_cookies'):
                Context.cookies = get_cookie_string_from(client.cookies)

            context_vars = _.get('req').get('context', {})
            for var_name, element in context_vars.items():
                if re.match(r'^(.*)\[\]$', var_name):
                    var_name = re.match(r'^(.*)\[\]$', var_name)[1]
                    if Context.get(var_name):
                        Context.get(var_name).append(get_attribute_value_from_json(response, element))
                    else:
                        empty_list = []
                        empty_list.append(get_attribute_value_from_json(response, element))
                        setattr(Context, var_name, empty_list)
                else:
                    setattr(Context, var_name, get_attribute_value_from_json(response, element))
        else:
            raise Exception('default data generation failed: ' + url + '\n\n' + str(_.get('req')) + '\n\n' + str(response))

        if _.get('req').get('wait'):
            time.sleep(int(_.get('req').get('wait')))
        

    return response


def form_endpoint(endpoint, request_body):
    if re.match(r'(.*/)?(<.*>)(/.*)?', endpoint):
        for k, v in request_body.items():
            endpoint = endpoint.replace('<' + k + '>', str(v))

    return endpoint


def get_cookie_string_from(cookie_jar):
    cookie_dict = cookie_jar.get_dict()
    cookie_string = ''
    for k,v in cookie_dict.items():
        cookie_string = cookie_string + k + '=' + v + '; '
    
    return cookie_string[:-2]


def replace_context_var(obj):
    obj_ = copy.deepcopy(obj)
    for k,v in obj_.items():
        if re.match(r'Context\.(.*)', str(v)):
            context_variable = re.match(r'Context\.(.*)', str(v))[1]
            obj[k] = Context.get(context_variable)
    return obj


def set_default_data_to_context():
    yml_files = os.listdir(settings.TEST_DATA_PATH + '/yml/')
    for f in yml_files:
        data = get_data_from_yml(f)
        Context.update(data)



Context = CustomDict()




from django.core.management.base import BaseCommand
import re
import os
from django.conf import settings
import pkg_resources


class Command(BaseCommand):
    help = 'Creates test automatically.'

    def create_dir(self, directory):
        if not os.path.exists(directory):
            os.makedirs(directory)

    def create_file(self, file):
        open(file, 'a+')

    def update_file(self, file, content):
        f = open(file, 'a')
        f.write(content)

    def create_skeleton(self):
        # self.create_dir(settings.BASE_DIR + '/' + self.app_name + '/tests')
        self.create_file(settings.BASE_DIR + '/' + self.app_name + '/test_unit.py')

    def write_to_unit_tests(self, content):
        with open(settings.BASE_DIR + '/' + self.app_name + '/test_unit.py', 'a') as f:
            f.write(content)

    def write_to_unit_test_imports(self, content):
        with open(settings.BASE_DIR + '/utils/test/' + 'unittest.py',
                  'a') as f:
            f.write(content)

    def create_test_data_file(self, test_method):
        if not os.path.exists(settings.BASE_DIR + '/utils/test/data/unittest/' + test_method + '.csv'):
            os.mknod(settings.BASE_DIR + '/utils/test/data/unittest/' + test_method + '.csv')

    def get_template(self, file_name):
        template = open(self.templates_dir + file_name, 'r')
        return template.read()

    def create_unit_test(self,
                         test_name,
                         import_path,
                         test_class=None,
                         static_method=False):

        try:
            current_tests = open(settings.BASE_DIR + '/' + self.app_name + '/test_unit.py', 'r').read()
            test_method_template = self.get_template(
                'unit_test_method_template.txt')
            test_import_template = self.get_template(
                'unit_test_import_template.txt')
            self.create_test_data_file(test_name)

            if current_tests:
                test_method_template = test_method_template.split('<><><><><><><><><>')[1]
            else:
                test_method_template = test_method_template.replace('<><><><><><><><><>', '')

            test_method_template = test_method_template.replace(
                '<test_name>', test_name)

            if test_class:
                test_import_template = test_import_template.replace(
                    '<test_object>', test_class)
                test_method_template = test_method_template.replace(
                    '<test_class_name>', test_class)
            else:
                test_method_template = test_method_template.replace(
                    '\'<test_class_name>\'', 'False')
                test_import_template = test_import_template.replace(
                    '<test_object>', test_name)
            
            if not static_method:
                test_method_template = test_method_template.replace(
                '<static>', 'False')
            else:
                test_method_template = test_method_template.replace(
                '<static>', 'True')
            

            test_method_template = test_method_template.replace(
                '<test_method_name>', test_name)
            test_import_template = test_import_template.replace(
                'import_path', import_path)

            self.write_to_unit_tests(test_method_template)
            self.write_to_unit_test_imports(test_import_template)

            return True

        except Exception as e:
            print(e)
            return False

    def add_arguments(self, parser):
        parser.add_argument(
            '-t',
            default=None,
            type=str,
            help='Specify test name',
            required=True)
        parser.add_argument(
            '-a',
            default=None,
            type=str,
            help='Specify app name',
            required=True)
        parser.add_argument(
            '-p',
            default=None,
            type=str,
            help='Specify test method location',
            required=True)
        parser.add_argument(
            '-c', default=None, type=str, help='Specify test class')
        parser.add_argument(
            '--static',
            action='store_true',
            dest='static_method',
            help='Specifies your method as static',
        )

    def handle(self, *args, **options):
        self.templates_dir = pkg_resources.resource_filename('cmtestrunner', 'management/templates/')
        test_name = options.get('t')
        self.app_name = options.get('a')
        import_path = options.get('p')
        test_class = options.get('c', None)
        static_method = options.get('static_method', False)
        self.create_skeleton()

        if self.create_unit_test(test_name, import_path, test_class,
                                 static_method):
            self.stdout.write(
                self.style.SUCCESS('Successfully added test "%s"' % test_name))
        else:
            self.stdout.write(
                self.style.ERROR('Error while adding test: "%s"' % test_name))




import json
import copy
import re
import collections



class ObjectManager():

    def __init__(self, obj_x, obj_y, modifier_str=None):
        self.obj_1 = copy.deepcopy(obj_x)
        self.obj_2 = copy.deepcopy(obj_y)
        if modifier_str:
            self.modifier_str = modifier_str
        else:
            self.modifier_str = "<span class='text-danger'><<replace-here>></span>"
                
        self.matched = True
        self.mismatch_keys = []


    def match_schema(self, val_1, val_2):
        # match datetime +/- offset
        # match time +/- offset
        # match random token(regex)
        # match any value exists
        regex_found = re.match(r'REGEX\((.*)\)', str(val_2))
        
        if regex_found:
            regex = regex_found[1]
            return re.match(r'{}'.format(regex), str(val_1)) is not None




    def match_dict_obj(self, dict_1, dict_2):
        if dict_1 == dict_2:
            return True
        
        keys = list(set(dict_1.keys()) | set(dict_2.keys()))
        
        for key in keys:
            if dict_1.get(key) != dict_2.get(key):          
                if type(dict_1.get(key)) is list and type(dict_2.get(key)) is list:
                    self.match_list_obj(dict_1.get(key), dict_2.get(key))
                elif type(dict_1.get(key)) is dict and type(dict_2.get(key)) is dict:
                    self.match_dict_obj(dict_1.get(key), dict_2.get(key))
                elif self.match_schema(dict_1.get(key), dict_2.get(key)):
                    continue
                else:
                    self.mismatch_keys.append(key)
                    self.matched = False
                    dict_1[key] = self.modifier_str.replace('<<replace-here>>', str(dict_1.get(key)))
                    dict_2[key] = self.modifier_str.replace('<<replace-here>>', str(dict_2.get(key)))
                    # dict_1[key] = "<span class='text-danger'>" + str(dict_1.get(key)) + '</span>'
                    # dict_2[key] = "<span class='text-danger'>" + str(dict_2.get(key)) + '</span>'


    
    def match_list_obj(self, list_1, list_2):

        if list_1 == list_2:
            return True

        if len(list_2) > len(list_1):
            list_1 += [None] * (len(list_2) - len(list_1))
        if len(list_1) > len(list_2):
            list_2 += [None] * (len(list_1) - len(list_2))
        
        for index, element in enumerate(list_1):
            if element != list_2[index]:
                if type(element) is dict and type(list_2[index]) is dict:
                    self.match_dict_obj(element, list_2[index])
                elif type(element) is list and type(list_2[index]) is list:
                    self.match_list_obj(element, list_2[index])
                elif self.match_schema(element, list_2[index]):
                    continue
                else:
                    self.matched = False
                    list_1[index] = "<span class='text-danger'>" + str(list_1[index]) + '</span>'
                    list_2[index] = "<span class='text-danger'>" + str(list_2[index]) + '</span>'


    def match_obj(self):

        if self.obj_1 == self.obj_2:
            # self.matched = True
            return True

        if type(self.obj_1) is not type(self.obj_2):
            self.mismatch_keys.append('whole object mismatch')
            print('Can not compare two different type of objects')
            return False

        if not self.obj_1 or not self.obj_2:
            # self.mismatches.append([obj_1, obj_2])
            return False


        if type(self.obj_1) is list:
            self.match_list_obj(self.obj_1, self.obj_2)

        if type(self.obj_2) is dict:
            self.obj_1 = dict(collections.OrderedDict(sorted(self.obj_1.items())))
            self.obj_2 = dict(collections.OrderedDict(sorted(self.obj_2.items())))
            self.match_dict_obj(self.obj_1, self.obj_2)





    def get_converted(self):
        return self.obj_1, self.obj_2

    def is_matched(self):
        return self.matched

    def mismatched_keys(self):
        return list(set(self.mismatch_keys))


# def test_object_manager():
#     obj_2 = {
#         'a': 'test',
#         'b': 'REGEX([0-9]+)',
#         'c': 'REGEX(^[0-9]$)',
#     }

#     obj_1 = {
#         'a': 'test',
#         'b': 23432,
#         'c': 2222
#     }
#     obj_manager = ObjectManager(obj_1, obj_2)
#     obj_manager.match_obj()
#     print(obj_manager.get_converted())
#     print(obj_manager.is_matched())


# test_object_manager()


from .test_runner import TestRunner, CMTestRunner
from .middleware import (parse_list_string, request_response_formatter, 
                        process_request_response, set_lang_header, set_auth_header, 
                        get_random_string, add_str_to_each_list_element,
                        get_test_endpoints, get_int_value_if_available, dict_to_obj,
                        Context, create_default_data_by_api, get_data_from_yml)



from unittest import TestCase
from rest_framework.test import APIClient
from django.utils.translation import ugettext as _
from django.utils import translation
from .middleware import (
                            request_response_formatter, get_all_locales,
                            get_translations, reset_db, BOOL,
                            set_auth_header, reset_auth_header, set_custom_headers,
                            set_lang_header, set_reset_seq_query, set_all_models,
                            unit_test_formatter, generate_test_report,
                            parse_snapshot, generate_analytics, get_test_endpoints,
                            Constants, mark_test_as_failed, form_endpoint, replace_context_var,
                            set_default_data_to_context
                        )
from unittest import TextTestRunner, TextTestResult
from django.test.runner import DiscoverRunner
import inspect
import traceback
from rest_framework.test import RequestsClient
from django.conf import settings
import requests
import re
import json
import os
import csv
from django.template.loader import render_to_string
from .object_manager import ObjectManager
import time
import copy


class CustomTextTestResult(TextTestResult):
    def startTestRun(self):
        set_reset_seq_query()
        set_all_models()
        set_default_data_to_context()

    def stopTestRun(self):
        self.testsRun = TestRunner.total_test_cases
        analytics = generate_analytics(Constants.FAIL_LOG)
        if TestRunner.query_executor:
            TestRunner.query_executor.quit()

        rendered = render_to_string('report.html', {
            'priority_fail_count': analytics,
            'passed_tests': Constants.PASSED_TESTS,
            'failed_tests': Constants.FAILED_TESTS,
            'success_count': self.testsRun - len(Constants.FAIL_LOG),
            'success_percentage': float("%0.2f"%((self.testsRun - len(Constants.FAIL_LOG))/(self.testsRun + len(Constants.EXCEPTIONS)) * 100)),
            'fail_count': len(Constants.FAIL_LOG),
            'fail_percentage': float("%0.2f"%(len(Constants.FAIL_LOG)/(self.testsRun + len(Constants.EXCEPTIONS)) * 100)),
            'exception_count': len(Constants.EXCEPTIONS),
            'exception_percentage': float("%0.2f"%(len(Constants.EXCEPTIONS)/(self.testsRun + len(Constants.EXCEPTIONS)) * 100)),
            'exception_details': Constants.EXCEPTIONS
        })

        
        if not os.path.exists(settings.TEST_PAYLOAD_PATH + '/reports'):
            os.makedirs(settings.TEST_PAYLOAD_PATH + '/reports')

        f = open(settings.TEST_PAYLOAD_PATH + '/reports/report.html', 'w')
        f.write(rendered)


    def startTest(self, test):
        "Called when the given test is about to be run"
        self._mirrorOutput = False
        self._setupStdout()


class CustomTextTestRunner(TextTestRunner):
    resultclass = CustomTextTestResult


class CMTestRunner(DiscoverRunner):
    test_runner = CustomTextTestRunner


class TestRunner(TestCase):
    client = False
    total_test_cases = 0
    create_default_user = lambda:None
    create_default_superuser = lambda:None
    create_default_superuser_token = lambda:None
    create_default_user_token = lambda:None
    reset_db = lambda:None
    total_exceptions = 0
    package = None
    ENDPOINTS = None
    query_executor = None



    def set_test_vars(self, test_data):
        # TestRunner.total_test_cases += 1
        self.response = None
        reset_auth_header(TestRunner.client)
        self.request_body = test_data.get('req')
        if self.request_body.get('user_type'):
            user_type = self.request_body.pop('user_type')
            if user_type == 'user':
                self.set_user_auth()
            elif user_type == 'admin':
                self.set_superuser_auth()
            elif user_type == 'invalid':
                set_auth_header(TestRunner.client, 'XXXinvalid_tokenXXX')
     
        self.reset_env = self.request_body.get('reset_env')
        

        self.wait = self.request_body.get('wait', 0)
        self.exp_response = test_data.get('resp')
        self.priority = self.request_body.pop('priority')
        self.test_id = self.request_body.pop('test_id')
        self.test_purpose = self.request_body.pop('comment')
        self.error_info = '\nTest ID: ' + str(
            self.test_id) + ' =====>>>>> ' + self.test_data_set + \
            ' failed for language: %s.\n' % self.accept_lang
        self.custom_headers = test_data.get('headers')
        # set_custom_headers(TestRunner.client, self.custom_headers)

        # if re.match(r'(.*/)?(<.*>)(/.*)?', self.endpoint):
        #     for k, v in self.request_body.items():
        #         self.endpoint = self.endpoint.replace('<' + k + '>', str(v))
        #         setattr(TestRunner.ENDPOINTS, self.endpoint_alias, self.endpoint)
        self.endpoint = form_endpoint(self.endpoint, self.request_body)
        setattr(TestRunner.ENDPOINTS, self.endpoint_alias, self.endpoint)


    def get_error(self, error_info, exp_response):
        errors = 'Validation failed for: ' + error_info + '\n\n'
        errors += 'Test Purpose: ' + self.test_purpose + '\n\n'
        errors += 'Requset Header: ' + str(
            TestRunner.client.headers) + '\n\n'
        errors += 'Request Body: \n\t' + str(self.request_body)[:100] + (
            ['', '....}'][len(str(self.request_body)) > 100]) + '\n\n'
        return errors

    def format_expected_response(self, exp_resp, accept_lang):
        translations = self.load_translations(accept_lang)
        if type(exp_resp) is list:
            for idx, x in enumerate(exp_resp):
                if type(x) is not dict:
                    if translations.get(x):
                        exp_resp[idx] = translations.get(x)
                    elif type(x) is not int:
                        exp_resp[idx] = _(x)
        elif type(exp_resp) is not dict:
            if translations.get(exp_resp):
                exp_resp = translations.get(exp_resp)
            elif type(exp_resp) is not int and type(exp_resp) is not float:
                exp_resp = _(exp_resp)
        return exp_resp

    def load_translations(self, accept_lang):
        return get_translations(accept_lang)


    def set_endpoints(self):
        TestRunner.ENDPOINTS = get_test_endpoints('endpoints.yml')

    def set_environment(self, envs):
        settings.TEST_SERVER = getattr(settings, self.package.upper() + '_BASE_URL')
        self.set_endpoints()
        TestRunner.client = self.get_client()   # need to find out why this is here again
        if settings.TEST_SERVER == 'http://testserver':
            TestRunner.reset_db = reset_db
        if TestRunner.query_executor:
            TestRunner.reset_db(TestRunner.query_executor)
        translation.activate('en')
        self.reproduce_steps = []
        try:
            for env in envs:
                self.reproduce_steps.append(env.__doc__)
                args = inspect.getargspec(env).args
                if args and args[0] == 'client':
                    env(TestRunner.client)
                else:
                    env()
        except Exception as e:
            Constants.EXCEPTIONS.append({
                'test_method': self.test_method.__name__,
                'details': e
            })
            raise
            # raise Exception(env.__doc__ + ':: ' + str(e))


    def set_headers(self, **kwargs):
        accept_lang = kwargs.get('accept_lang')
        set_lang_header(TestRunner.client, accept_lang)
        return True

    def set_user_auth(self):
        TestRunner.create_default_user()
        token = TestRunner.create_default_user_token()
        set_auth_header(TestRunner.client, token)

    def set_superuser_auth(self):
        TestRunner.create_default_superuser()
        token = TestRunner.create_default_superuser_token()
        set_auth_header(TestRunner.client, token)

    def get_response(self):
        return self.response

    def get_client(self):
        if settings.TEST_SERVER == 'http://testserver':
            return RequestsClient()
        else:
            return requests.Session()

    # see CONTRIBUTING.md for test data formats.
    def verify_test_result(self, exp_response, test_id, accept_lang):

        response_ = {}
        response_time = self.response.pop('response_time')

        

        exp_response_ = copy.deepcopy(exp_response)

        for key, value in exp_response_.items():
            if accept_lang != 'en':
                exp_response[key] = self.format_expected_response(
                    exp_response[key], accept_lang)
            if key == 'response':
                exp_response.pop('response')
                exp_response.update(parse_snapshot(value, self.response))   # there is a bug in this line, response may not be a dictionary always
                response_.update(self.response)
                continue
            exp_response[key] = value = parse_snapshot(value, self.response.get(key))
            response_[key] = self.response.get(key)
            exp_response = replace_context_var(exp_response)
        
        self.response = response_

        
        object_manager = ObjectManager(self.response, exp_response)
        object_manager.match_obj()
        self.response, exp_response = object_manager.get_converted()
        self.is_matched = object_manager.is_matched()
        error_info = ', '.join(object_manager.mismatched_keys())
        errors = self.get_error(error_info, exp_response)
        
        generate_test_report(
                test_name=self.test_data_set, priority=self.priority, test_id=self.test_id, 
                purpose=self.test_purpose, reproduce_steps=self.reproduce_steps,
                request_body=self.request_body, response=self.response, 
                request_header=dict(TestRunner.client.headers),
                expected_response=exp_response,
                error_info=error_info,
                response_time=response_time
                )
        
        try:
            self.assertEqual(self.is_matched, True, msg=errors)
        except AssertionError:
            mark_test_as_failed()
            raise


    def set_test_attributes(self, **kwargs):
        # TestRunner.client = self.get_client()
        self.test_method = kwargs.get('test_method')
        self.environment = kwargs.get('env', [])
        self.set_environment(self.environment)
        self.sub_test = kwargs.get('sub_test')
        self.accept_lang = kwargs.get('accept_lang')
        self.set_headers(accept_lang=self.accept_lang)
        self.test_data_set = kwargs.get('test_data_set')
        self.test_data = request_response_formatter('tests/' + self.test_data_set)
        self.endpoint_alias = kwargs.get('test_method').__name__.upper() + '_URL'
        self.endpoint = getattr(
            TestRunner.ENDPOINTS, 
            self.endpoint_alias
            )
        



    def process_tests(self, **kwargs):
        
        self.set_test_attributes(**kwargs)
            
        for each_test_data in self.test_data:
            self.set_test_vars(each_test_data)

            if self.request_body.get(
                    'accept_lang'
            ) and self.request_body.get('accept_lang') != self.accept_lang:
                TestRunner.total_test_cases -= 1
                continue
            time.sleep(float(self.wait)/1000)
            if self.reset_env:
                self.set_environment(self.environment)
            
        
            files = self.request_body.pop('files')
            self.request_body = replace_context_var(self.request_body)
            self.request_body['files'] = files
            set_custom_headers(TestRunner.client, replace_context_var(self.custom_headers))
            
            try:
                self.response = kwargs.get('test_method')(
                    client=TestRunner.client,
                    request_body=self.request_body,
                    accept_lang=self.accept_lang,
                    )
                
            except Exception as e:
                print(traceback.format_exc())
                errors = self.get_error(self.error_info, self.exp_response)
                with self.subTest():
                    self.assertEqual(None, e, msg=errors)
            with self.subTest():
                TestRunner.total_test_cases += 1
                self.verify_test_result(self.exp_response, self.test_id,
                                        self.accept_lang)

    def execute_tests(self, **kwargs):
        locales = get_all_locales()
        locales.append('en')
        for each_lang in locales:
            kwargs['accept_lang'] = each_lang
            self.process_tests(**kwargs)

    def execute_unit_tests(self, **kwargs):

        test_data_set = kwargs.get('test_data_set')
        test_data = unit_test_formatter(test_data_set)
        self.set_environment(kwargs.get('env'))

        for each_test_data in test_data:
            args = each_test_data.get('args')
            inits = each_test_data.get('inits')
            kwargs_ = each_test_data.get('kwargs')
            returns = each_test_data.get('returns')
            test_obj = kwargs.get('test_object')
            if kwargs.get('test_class'):
                test_class = getattr(test_obj, kwargs.get('test_class'))
                test_method = kwargs.get('test_method')
                if kwargs.get('static'):
                    test_obj = getattr(test_class, test_method)
                else:
                    test_obj = test_class(*inits)
                    test_obj = getattr(test_obj, test_method)
            else:
                test_obj = getattr(test_obj, kwargs.get('test_method'))
            response = test_obj(*args, **kwargs_)
            TestRunner.total_test_cases += 1

            self.assertEqual(str(response), str(returns))



from django.core.management.base import BaseCommand
import re
import os
from django.conf import settings
import pkg_resources

class Command(BaseCommand):
    help = 'creates test skeleton for API testing'

    def create_dir(self, directory):
        if not os.path.exists(directory):
            os.makedirs(directory)
            print('created ' + directory)

    def create_file(self, file):
        open(file, 'a+')
        print('created ' + file)

    def handle(self, *args, **kwargs):
        self.create_dir(settings.BASE_DIR + '/utils')
        self.create_dir(settings.BASE_DIR + '/utils/test')
        self.create_dir(settings.BASE_DIR + '/utils/test/reports')
        self.create_dir(settings.BASE_DIR + '/utils/test/data')
        self.create_dir(settings.BASE_DIR + '/utils/test/data/unittest')
        self.create_dir(settings.BASE_DIR + '/utils/test/data/snapshots')
        self.create_dir(settings.BASE_DIR + '/utils/test/data/upload')
        self.create_dir(settings.BASE_DIR + '/utils/test/data/tests')
        self.create_dir(settings.BASE_DIR + '/utils/test/data/default')
        self.create_dir(settings.BASE_DIR + '/utils/test/data/yml')
        self.create_dir(settings.BASE_DIR + '/utils/test/translation')
        self.create_dir(settings.BASE_DIR + '/utils/fixtures')

        self.create_file(settings.BASE_DIR + '/utils/__init__.py')
        self.create_file(settings.BASE_DIR + '/utils/test/endpoints.yml')
        self.create_file(settings.BASE_DIR + '/utils/test/unittest.py')
        self.create_file(settings.BASE_DIR + '/utils/test/test_runner.py')
        self.create_file(settings.BASE_DIR + '/utils/test/middleware.py')

        templates_dir = pkg_resources.resource_filename('cmtestrunner', 'management/templates/')
        test_runner_content = None
        with open(templates_dir + 'test_runner_template.txt', 'r') as f:
            test_runner_content = f.read()

        with open(settings.BASE_DIR + '/utils/test/test_runner.py' , 'w') as f:
            f.write(test_runner_content)


        self.stdout.write(
                self.style.SUCCESS('Successfully created test skeleton'))



import setuptools

setuptools.setup(name='cmtestrunner',
      version='2.1.2',
      description='custom test runner for API test',
      author='Tasin Nawaz',
      author_email='tasin.buet@gmail.com',
      license='CM',
      url='https://github.com/tasin-megamind/cmtestrunner',
      packages=setuptools.find_packages(),
      include_package_data=True,
      install_requires=[
          'django>=2.0.6',
          'djangorestframework>=3.8.2',
          'PyYAML>=3.12',
          'requests>=2.20.1',
          'numpy>=1.14.1',
      ],
      zip_safe=False)




from django.core.management.base import BaseCommand
import re
import os
from django.conf import settings
import pkg_resources

class Command(BaseCommand):
    help = 'creates test'

    def create_dir(self, directory):
        if not os.path.exists(directory):
            os.makedirs(directory)

    def create_file(self, file):
        open(file, 'a+')

    def update_file(self, file, content):
        f = open(file, 'a')
        f.write(content)

    def get_template(self, file):
        f = open(self.templates_dir + file, 'r')
        return f.read()

    def add_arguments(self, parser):
        parser.add_argument(
            '-t',
            default=None,
            type=str,
            help='Specify test name',
            required=True)
        parser.add_argument(
            '-X',
            default=None,
            type=str,
            help='Specify request method',
            required=True)
        parser.add_argument(
            '-u', 
            default=None, 
            type=str, 
            help='Specify test endpoint withot host & port')
        parser.add_argument(
            '-a',
            default=None,
            type=str,
            help='Specify app name',
            required=True)

    def generate_tests(self):
        tests_template = self.get_template('tests_template.txt')
        current_tests = open(self.base_dir + '/' + self.app_name + '/test_api.py', 'r').read()

        tests_template = tests_template.replace('<test_name>', self.test_name)
        if re.sub(r'#.*', '', current_tests).strip():
            tests_template = tests_template.split('<><><><><><><><><><>')[1]    
        else:
            tests_template = tests_template.replace('<><><><><><><><><><>', '')
            tests_template = tests_template.replace('<app>', self.app_name)
            tests_template = tests_template.replace('<App>', self.app_name.capitalize())

        self.update_file(self.base_dir + '/' + self.app_name + '/test_api.py', tests_template)
    
    def generate_test_methods(self):
        test_methods_template = self.get_template('test_methods_template.txt')
        current_test_methods = open(self.base_dir + '/' + self.app_name + '/test_methods.py', 'r').read()

        test_methods_template = test_methods_template.replace('<test_name>', self.test_name)
        test_methods_template = test_methods_template.replace('<request_method>', self.request_method)
        test_methods_template = test_methods_template.replace('<test_url_alias>', self.test_url_alias)
        if re.sub(r'#.*', '', current_test_methods).strip():
            test_methods_template = test_methods_template.split('<><><><><><><><><><>')[1]
        else:
            test_methods_template = test_methods_template.replace('<><><><><><><><><><>', '')
        
        self.update_file(self.base_dir + '/' + self.app_name + '/test_methods.py', test_methods_template)

    def generate_endpoints(self):
        endpoints_template = self.get_template('endpoints_template.txt')
        endpoints_template = endpoints_template.replace('<test_url_alias>', self.test_url_alias)
        endpoints_template = endpoints_template.replace('<test_url>', self.test_url)
        self.update_file(self.base_dir + '/utils/test/endpoints.yml', endpoints_template)

    def handle(self, *args, **kwargs):
        self.templates_dir = pkg_resources.resource_filename('cmtestrunner', 'management/templates/')
        self.app_name = kwargs.get('a')
        self.test_url = kwargs.get('u')
        self.request_method = kwargs.get('X')
        self.test_name = kwargs.get('t')
        self.base_dir = settings.BASE_DIR

        self.test_url_alias = self.test_name.upper() + '_URL'
        # self.create_dir(self.base_dir + '/' + self.app_name + '/tests')
        self.create_file(self.base_dir + '/' + self.app_name + '/test_api.py')
        self.create_file(self.base_dir + '/' + self.app_name + '/test_methods.py')

        self.generate_endpoints()
        self.generate_test_methods()
        self.generate_tests()
        
        if not os.path.exists(self.base_dir + '/utils/test/data/tests/' + self.test_name + '.csv'):
            file_content = 'test_id,request_body,resp_response,priority,comment'
            self.update_file(self.base_dir + '/utils/test/data/tests/' + self.test_name + '.csv', file_content)

        self.stdout.write(
                self.style.SUCCESS('Successfully created test'))
        











